from abc import abstractmethod
from pathlib import Path
from typing import Any, Iterable, Type

import pytest

from pytest_logikal.core import ReportInfoType


class ItemRunError(Exception):
    """An item run error."""


class Item(pytest.Item):
    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.add_marker(self.name)

    def reportinfo(self) -> ReportInfoType:
        info = f'[{self.name}]'
        if self.path:
            relative_path = self.path.relative_to(self.config.invocation_params.dir)
            info += f' {relative_path}'
        return (self.path, None, info)

    def repr_failure(
        self, excinfo: pytest.ExceptionInfo[BaseException], *args: Any, **kwargs: Any,
    ) -> Any:
        if excinfo.errisinstance(ItemRunError):
            return str(excinfo.value)
        return super().repr_failure(excinfo, *args, **kwargs)

    @abstractmethod
    def runtest(self) -> None:
        ...


class Plugin:
    name: str
    item: Type[Item]

    def __init__(self, config: pytest.Config):
        if not hasattr(self, 'name'):
            raise AttributeError('You must provide a plugin name')
        if not hasattr(self, 'item'):
            raise AttributeError('You must provide an item type')

        self.config = config

    def pytest_collect_file(self, parent: pytest.Collector) -> Any:
        plugin = self

        class File(pytest.File):
            def collect(self) -> Iterable[Item]:
                # Ensure that a plugin item is added once and only once per session
                if not any(isinstance(item, plugin.item) for item in self.session.items):
                    yield plugin.item.from_parent(parent=self, name=plugin.name)
        return File.from_parent(parent, path=Path('check'))



import json
import subprocess

import pytest

from pytest_logikal.core import PYPROJECT
from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('pylint')
    group.addoption('--pylint', action='store_true', default=False, help='run pylint')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'pylint: tests linting source code.')
    if config.option.pylint:
        config.pluginmanager.register(PylintPlugin(config=config))


class PylintItem(CachedFileCheckItem):
    def run(self) -> None:
        plugins = [
            'pylint.extensions.code_style',
            'pylint.extensions.emptystring',
            'pylint.extensions.comparetozero',
            'pylint.extensions.comparison_placement',
            'pylint.extensions.confusing_elif',
            'pylint.extensions.for_any_all',
            'pylint.extensions.consider_refactoring_into_while_condition',
            'pylint.extensions.consider_ternary_expression',
            'pylint.extensions.bad_builtin',
            'pylint.extensions.mccabe',
            'pylint.extensions.dict_init_mutate',
            'pylint.extensions.dunder',
            'pylint.extensions.check_elif',
            'pylint.extensions.empty_comment',
            'pylint.extensions.eq_without_hash',
            'pylint.extensions.private_import',
            'pylint.extensions.no_self_use',
            'pylint.extensions.overlapping_exceptions',
            'pylint.extensions.redefined_loop_name',
            'pylint.extensions.set_membership',
            'pylint.extensions.typing',
            'pylint.extensions.while_used',
        ]
        command = [
            'pylint', str(self.path),
            '--init-hook=import sys; sys.path.append(".")',
            f'--max-line-length={self.config.getini("max_line_length")}',
            '--include-naming-hint=y',
            '--output-format=json',
            f'--max-complexity={self.config.getini("max_complexity")}',
        ]
        enable = ['useless-suppression', 'use-symbolic-message-instead']
        disable = [
            # checks covered by pycodestyle
            'line-too-long', 'trailing-newlines', 'trailing-whitespace', 'missing-final-newline',
            # other checks
            'duplicate-code',  # not working with distributed exeuction
            'logging-fstring-interpolation',  # we are mostly using f-strings in logging
            'missing-docstring',  # we are less strict about class and function docstrings
            'consider-using-tuple',  # lists are often easier to read
            'use-dict-literal',  # the dict class constructor approach is sometimes useful
            'consider-using-namedtuple-or-dataclass',  # unnamed tuples can be sometimes useful
        ]
        if 'DJANGO_SETTINGS_MODULE' in self.config.inicfg:
            disable += [
                'too-few-public-methods',  # common error with some Django classes
                'unsubscriptable-object',  # common error with generic types in django-stubs
            ]
            plugins += ['pylint_django']
            command += [
                f'--django-settings-module={self.config.inicfg["DJANGO_SETTINGS_MODULE"]}',
                r'--module-rgx=[^\WA-Z]*$',  # allow (migration) modules to start with digits
            ]

        pyproject_pylint = PYPROJECT.get('tool', {}).get('pylint', {}).get('messages_control', {})
        enable = pyproject_pylint.get('enable', enable)
        disable = pyproject_pylint.get('disable', disable)
        command += [
            f'--enable={",".join(enable)}', f'--disable={",".join(disable)}',
            f'--load-plugins={",".join(plugins)}',
        ]

        # Note that we are running Pylint in a subprocess and process its output instead of
        # importing it due to its license (GPLv2). The subprocess call is secure as it is not using
        # untrusted input.
        process = subprocess.run(command, capture_output=True, text=True, check=False)  # nosec
        try:
            if messages := json.loads(process.stdout):
                formatter = '{line}:{column}: {type}: {message} ({symbol})'
                raise ItemRunError('\n'.join(formatter.format(**message) for message in messages))
        except json.decoder.JSONDecodeError as error:
            raise ItemRunError(f'Error: {process.stdout or process.stderr}') from error


class PylintPlugin(CachedFileCheckPlugin):
    name = 'pylint'
    item = PylintItem



from itertools import permutations
from pathlib import Path
from typing import List, Optional, Tuple, Union

import pytest
from babel.messages import Catalog, Message, mofile, pofile

from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('translations')
    group.addoption('--translations', action='store_true', default=False,
                    help='run django translation checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'translations: checks Django translation files.')
    if config.option.translations:
        config.pluginmanager.register(TranslationPlugin(config=config))


def _get_message_context(message: Message) -> Optional[Union[str, bytes]]:
    if isinstance(message.context, bytes):
        return message.context.decode('utf-8')  # type: ignore[unreachable]
    if isinstance(message.context, str):
        return message.context.encode()
    return message.context


def _get_message_string(message: Message) -> Union[str, Tuple[str]]:
    if isinstance(message.string, list):
        return tuple(message.string)  # type: ignore[return-value]
    return message.string  # type: ignore[return-value]


def _message_in_catalog(message: Message, catalog: Catalog) -> bool:
    catalog_message = catalog.get(
        id=message.id,
        context=_get_message_context(message),  # type: ignore[arg-type]
    )
    return (
        catalog_message is not None
        and _get_message_string(message) == _get_message_string(catalog_message)
        and message.flags == catalog_message.flags
    )


def _catalogs_identical(catalog: Catalog, compiled_catalog: Catalog) -> bool:
    if dict(catalog.mime_headers) != dict(compiled_catalog.mime_headers):
        return False
    for catalog_1, catalog_2 in permutations([catalog, compiled_catalog]):
        for message in catalog_1:
            if not message.id:
                continue
            if not _message_in_catalog(message, catalog_2):
                return False
    return True


class TranslationItem(CachedFileCheckItem):
    plugin: 'TranslationPlugin'

    def run(self) -> None:
        errors: List[str] = []

        with open(self.path, encoding='utf-8') as catalog_file:
            catalog = pofile.read_po(catalog_file, abort_invalid=True)

        # Check compiled translation file
        compiled_path = self.path.with_suffix('.mo')
        if compiled_path.exists():
            with open(compiled_path, mode='rb') as compiled_catalog_file:
                compiled_catalog = mofile.read_mo(compiled_catalog_file)
            if not _catalogs_identical(catalog, compiled_catalog):
                errors.append(f'error: Compiled translation file "{compiled_path}" is outdated')
        else:
            errors.append(f'error: Compiled translation file "{compiled_path}" does not exist')

        # Check catalog fuzziness
        if catalog.fuzzy:
            errors.append('error: Fuzzy catalog')

        # Check messages
        for message in catalog:
            # Check fuzziness
            if message.fuzzy and message.lineno:  # we skip the catalog fuzzy marker
                errors.append(f'{message.lineno}: error: Fuzzy message')

            # Check missing translation
            messages = message.string if isinstance(message.string, tuple) else [message.string]
            if any(not string for string in messages):
                errors.append(f'{message.lineno}: error: Missing translation')

        # Report errors
        if errors:
            raise ItemRunError('\n'.join(errors))


class TranslationPlugin(CachedFileCheckPlugin):
    name = 'translations'
    item = TranslationItem

    def check_file(self, file_path: Path) -> bool:
        return file_path.suffix == '.po'



import re
from io import StringIO
from typing import Any, Dict

import isort
import pytest
from isort.wrap_modes import WrapModes

from pytest_logikal.core import PYPROJECT
from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError


def get_config(max_line_length: int) -> Dict[str, Any]:
    config = {
        'py_version': 'auto',
        'line_length': max_line_length,
        'multi_line_output': WrapModes.VERTICAL_GRID_GROUPED,  # type: ignore[attr-defined]
        'balanced_wrapping': True,
        'combine_as_imports': True,
        'use_parentheses': True,
        'include_trailing_comma': True,
        'color_output': True,
    }
    if 'tool' in PYPROJECT and 'isort' in PYPROJECT['tool']:
        for option in config:
            if option in PYPROJECT['tool']['isort']:
                config[option] = PYPROJECT['tool']['isort'][option]
    return config


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('isort')
    group.addoption('--isort', action='store_true', default=False, help='run isort')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'isort: tests checking import ordering.')
    if config.option.isort:
        config.pluginmanager.register(IsortPlugin(config=config))


class IsortItem(CachedFileCheckItem):
    def run(self) -> None:
        config = get_config(max_line_length=int(self.config.getini('max_line_length')))
        stdout = StringIO()
        if not isort.check_file(filename=str(self.path), show_diff=stdout, **config):
            diff = stdout.getvalue()
            message = re.sub('^([+]{3}|[-]{3}|[@]{2}) .*\n', '', diff, flags=re.MULTILINE)
            raise ItemRunError(message.lstrip('\n').rstrip())


class IsortPlugin(CachedFileCheckPlugin):
    name = 'isort'
    item = IsortItem



# pylint: disable=import-outside-toplevel
import inspect
import logging
import re
from contextlib import contextmanager
from importlib import import_module
from operator import itemgetter
from pathlib import Path
from time import sleep
from typing import Any, Dict, Iterable, Iterator, Optional, TypeVar, Union, cast

import pytest
from selenium.common.exceptions import SessionNotCreatedException
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.chrome.webdriver import WebDriver as Chrome

from pytest_logikal.core import PYPROJECT
from pytest_logikal.utils import Fixture, Function, assert_image_equal

logger = logging.getLogger(__name__)

Setting = TypeVar('Setting')


class BrowserSettings:  # pylint: disable=too-many-instance-attributes
    UNLIMITED_HEIGHT = 'unlimited'
    FRAME_HEIGHT = 'frame'

    # Note: make sure to update the docstring of the set_browser fixture when changing the defaults
    DEFAULTS = {
        'desktop': {'width': 1800, 'frame_height': 900},  # 1920x1080 -dash -scrollbar -header
        'laptop-l': {'width': 1440, 'frame_height': 900},
        'laptop': {'width': 1024, 'frame_height': 768},
        'tablet': {'width': 768, 'frame_height': 1024, 'mobile': True},
        'mobile-l': {'width': 425, 'frame_height': 680, 'mobile': True},
        'mobile-m': {'width': 375, 'frame_height': 600, 'mobile': True},
        'mobile-s': {'width': 320, 'frame_height': 512, 'mobile': True},
    }

    def __init__(  # pylint: disable=too-many-arguments
        self,
        setting: Optional[str] = 'desktop',
        name: Optional[str] = None,
        width: Optional[int] = None,
        height: Optional[Union[int, str]] = None,
        frame_height: Optional[int] = None,
        mobile: Optional[bool] = None,
    ):
        """
        Browser settings data.

        Args:
            setting: The named setting to use.
            name: The name to use for screenshots. Defaults to the setting name.
            width: The width of the browser window.
            height: The height of the browser window (use 'unlimited' for unlimited and 'frame' to
                use the frame height). Defaults to 'unlimited'.
            frame_height: The browser window frame height to use when taking 'unlimited'
                screenshots.
            mobile: Whether it is a mobile browser. Defaults to :data:`False`.

        """
        pyproject_settings = PYPROJECT.get('tool', {}).get('browser', {}).get('settings', {})
        settings = {**self.DEFAULTS, **pyproject_settings}
        if setting and setting not in settings:
            raise ValueError(f'Unknown browser setting "{setting}"')
        self.settings = settings[setting] if setting else {}
        self.setting = setting
        self.name = name or setting
        self.width = self._setting(width, 'width')

        self.height = self._setting(height, 'height', self.UNLIMITED_HEIGHT)
        regular_height = self.height not in (self.UNLIMITED_HEIGHT, self.FRAME_HEIGHT)
        self.frame_height = self._setting(frame_height, 'frame_height',
                                          default=self.height if regular_height else None)
        self.browser_height = self.height if regular_height else self.frame_height

        self.mobile = self._setting(mobile, 'mobile', False)

    def _setting(
        self, value: Optional[Setting], name: str, default: Optional[Setting] = None,
    ) -> Setting:
        setting = value if value is not None else self.settings.get(name, default)  # type: ignore
        if default is None and setting is None:
            raise ValueError(f'The browser setting "{name}" must be specified')
        return setting  # type: ignore

    def __str__(self) -> str:
        attributes = ['setting', 'name', 'width', 'height', 'frame_height', 'mobile']
        settings = ', '.join(f'{name!r}: {getattr(self, name)!r}' for name in attributes)
        return f'{{{settings}}}'


class Browser:
    def __init__(
        self,
        temp_path: Path,
        settings: Optional[BrowserSettings] = None,
        screenshot_path: Path = Path('screenshot'),
    ):
        """
        Browser automation.

        Has the same attributes and methods as a :mod:`selenium.webdriver.Chrome
        <selenium.webdriver.chrome.webdriver>` instance.
        """
        self.temp_path = temp_path
        self.settings = settings or BrowserSettings()
        logger.debug(f'Browser settings: {self.settings}')
        self.name = self.settings.name
        self.screenshot_path = screenshot_path
        logger.debug(f'Using screenshot path "{self.screenshot_path}"')

        options = cast(Any, ChromeOptions())  # add_argument is not yet typed
        if self.settings.mobile:
            options.add_argument('--hide-scrollbars')
        options.add_argument('--headless')
        options.add_argument(f'--window-size={self.settings.width},{self.settings.browser_height}')
        self.driver = Browser._get_driver(options=options)

    @staticmethod
    def _get_driver(options: ChromeOptions) -> Chrome:
        try:
            return Chrome(options=options)
        except SessionNotCreatedException as error:
            if driver_version_error := Browser._driver_version_error(str(error.msg)):
                raise RuntimeError(driver_version_error) from None  # override parent exception
            raise

    @staticmethod
    def _driver_version_error(error_message: str) -> Optional[str]:
        search = re.search(
            'This version of ChromeDriver only supports Chrome version (.*?)\n'
            'Current browser version is (.*?) with binary path',
            error_message,
        )
        return '\n'.join([
            f'Invalid ChromeDriver version (compatible with Chrome {search.group(1)})',
            f'Please download a release compatible with Chrome version {search.group(2)} '
            '(see https://chromedriver.chromium.org/downloads)'
        ]) if search else None

    def __enter__(self) -> 'Browser':
        return self

    def __exit__(self, *args: Any) -> None:
        self.driver.__exit__(*args)

    def __getattr__(self, name: str) -> Any:
        return getattr(self.driver, name)

    @contextmanager
    def auto_height(self, wait_milliseconds: Optional[int]) -> Iterator[None]:
        if self.settings.height != BrowserSettings.UNLIMITED_HEIGHT:
            yield
            return
        logger.debug('Using full page height')
        original_size = self.driver.get_window_size()
        if wait_milliseconds:  # we use a small delay to mitigate height flakiness
            logger.debug(f'Waiting {wait_milliseconds} ms')
            sleep(wait_milliseconds / 1000)
        elements = [
            'document.body.clientHeight',
            'document.body.scrollHeight',
            'document.body.offsetHeight',
            'document.documentElement.clientHeight',
            'document.documentElement.scrollHeight',
            'document.documentElement.offsetHeight',
        ]
        script = f'return Math.max({",".join(elements)});'
        height = self.driver.execute_script(script)  # type: ignore[no-untyped-call]
        logger.debug(f'Calculated page height: {height}')
        self.driver.set_window_size(original_size['width'], height)
        try:
            yield
        finally:
            self.driver.set_window_size(**original_size)

    def check(self, name: Optional[str] = None, wait_milliseconds: Optional[int] = 100) -> None:
        """
        Create a screenshot and check it against an expected version.

        Args:
            name: The name of the expected screenshot.
            wait_milliseconds: The milliseconds to wait before calculating the screenshot height
                for unlimited height checks.

        """
        name_parts = [self.screenshot_path.name, self.name, name]
        full_name = '_'.join(part for part in name_parts if part is not None)
        expected = self.screenshot_path.with_name(full_name).with_suffix('.png')

        script = 'document.body.style.caretColor = "transparent";'  # hide the blinking caret
        self.driver.execute_script(script)  # type: ignore[no-untyped-call]

        with self.auto_height(wait_milliseconds=wait_milliseconds):
            logger.debug('Taking screenshot')
            # Note: we are disabling debug remote logs because they contain the verbose image data
            logging.getLogger('selenium.webdriver.remote').setLevel(logging.INFO)
            screenshot = self.driver.get_screenshot_as_png()
            logging.getLogger('selenium.webdriver.remote').setLevel(logging.DEBUG)
        assert_image_equal(actual=screenshot, expected=expected, temp_path=self.temp_path)

    def login(self, user: Any, force: bool = True) -> None:
        """
        .. note:: The ``django`` extra must be installed for this method to work.

        Log in a given user.

        Args:
            user: The user to log in.
            force: Whether to log the user in without going through the authentication steps.

        """
        try:
            from django.conf import settings
            from django.contrib.auth import login as django_auth_login
            from django.http import HttpRequest
        except ImportError as error:  # pragma: no cover
            raise RuntimeError('The `django` extra must be installed for login to work') from error
        if not force:
            raise NotImplementedError('Only the forced login is implemented currently')

        request = HttpRequest()
        request.session = import_module(settings.SESSION_ENGINE).SessionStore()
        django_auth_login(request, user)
        request.session.save()
        self.driver.add_cookie({
            'name': settings.SESSION_COOKIE_NAME,
            'value': request.session.session_key,
        })


@pytest.fixture
def browser(tmp_path: Path, request: Any) -> Iterable[Browser]:  # noqa: D400,D402,D415
    """
    browser() -> Browser

    Yield a :class:`Browser` instance.
    """
    function = request.node.obj
    if not (source_file := inspect.getsourcefile(function)):  # pragma: no cover, defensive line
        raise RuntimeError(f'Source file cannot be found for "{function}"')
    source_path = Path(source_file)
    screenshots = source_path.parent / 'screenshots' / source_path.stem / function.__name__

    settings = BrowserSettings(**getattr(request, 'param', {}))
    with Browser(temp_path=tmp_path, settings=settings, screenshot_path=screenshots) as instance:
        yield instance


def set_browser(*args: Dict[str, Any], **kwargs: Any) -> Fixture[Any]:
    """
    Control the creation of the :class:`Browser` object in the :func:`browser <browser>` fixture.

    Args:
        *args: A list of arguments to forward to the underlying :class:`BrowserSettings` instances.
        **kwargs: The arguments to forward to the underlying :class:`BrowserSettings` instance.

    You can specify browser settings with this decorator as follows:

    .. code-block:: python

        from pytest_django.live_server_helper import LiveServer
        from pytest_logikal.browser import Browser, set_browser


        @set_browser(setting='tablet')
        def test_homepage(browser: Browser, live_server: LiveServer) -> None:
            browser.get(live_server.url)
            browser.check()

    The currently available settings are the following:

    ========  =====  =========  =====  ======
    Setting   Width   Height    Frame  Mobile
    ========  =====  =========  =====  ======
    desktop   1800   unlimited    900  False
    laptop-l  1440   unlimited    900  False
    laptop    1024   unlimited    768  False
    tablet     768   unlimited   1024  True
    mobile-l   425   unlimited    680  True
    mobile-m   375   unlimited    600  True
    mobile-s   320   unlimited    512  True
    ========  =====  =========  =====  ======

    You can also run your test in multiple browser environments:

    .. code-block:: python

        @set_browser(
            dict(setting='desktop'),
            dict(setting='mobile'),
        )
        def test_homepage(browser: Browser, live_server: LiveServer) -> None:
            browser.get(live_server.url)
            browser.check()

    The named settings can be overridden or extended via the ``tool.browser.settings`` table in
    ``pyproject.toml``:

    .. code-block:: toml

        [tool.browser.settings]
        desktop = {width = 1024, height = 768}
        mobile-xl = {width = 480, frame_height = 768, mobile = true}

    You can then use your custom named setting as follows:

    .. code-block:: python

        @set_browser(setting='mobile-xl')
        def test_homepage(browser: Browser, live_server: LiveServer) -> None:
            browser.get(live_server.url)
            browser.check()

    While we recommend using named settings, you can also use completely arbitrary settings for a
    test as follows:

    .. code-block:: python

        @set_browser(name='laptop-xl', width=1920, height=1080)
        def test_homepage(browser: Browser, live_server: LiveServer) -> None:
            browser.get(live_server.url)
            browser.check()

    """
    argvalues = args or [kwargs]

    def test_id(argvalue: Dict[str, Any]) -> str:
        if 'name' in argvalue and (value := argvalue['name']):
            return str(value)
        if 'setting' in argvalue and len(argvalue) == 1 and (value := argvalue['setting']):
            return str(value)
        return re.sub(r'[{}\[\]\':,]', '', str(argvalue)).replace(' ', '_').lower()

    def parametrized_test_function(function: Function) -> Any:
        for argvalue in argvalues:
            argvalue['name'] = test_id(argvalue)
        return pytest.mark.parametrize(
            argnames='browser', argvalues=argvalues, indirect=True, ids=itemgetter('name'),
        )(function)

    return parametrized_test_function



from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
from typing import Optional

import pytest
from django.conf import settings
from django_migration_linter import MigrationLinter
from django_migration_linter.management.commands import lintmigrations
from pytest_django.plugin import _blocking_manager  # pylint: disable=import-private-name

from pytest_logikal.core import PYPROJECT, ReportInfoType
from pytest_logikal.plugin import Item, ItemRunError, Plugin


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('migration')
    group.addoption('--migration', action='store_true', default=False,
                    help='run django migration checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'migration: checks Django migrations.')
    if config.option.migration:
        config.pluginmanager.register(MigrationPlugin(config=config))


class MigrationItem(Item):
    def runtest(self, migrations_file_path: Optional[Path] = None) -> None:
        # We cannot request a fixture here, so we use the blocking manager directly
        # See https://github.com/pytest-dev/pytest/discussions/10915
        with _blocking_manager.unblock():
            linter = MigrationLinter(**{
                'all_warnings_as_errors': True,
                **getattr(settings, 'MIGRATION_LINTER_OPTIONS', {}),
                **PYPROJECT.get('tool', {}).get(lintmigrations.CONFIG_NAME, {}),
            })
            with redirect_stdout(StringIO()) as code_stdout:
                linter.lint_all_migrations(migrations_file_path=migrations_file_path)
            migrations = 'migration' if linter.nb_total == 1 else 'migrations'
            print(f'Checked {linter.nb_total} {migrations} ({linter.nb_ignored} ignored)')
            if linter.nb_warnings or linter.nb_erroneous:
                raise ItemRunError(code_stdout.getvalue().rstrip())

    def reportinfo(self) -> ReportInfoType:
        return (self.path, None, f'[{MigrationPlugin.name}]')


class MigrationPlugin(Plugin):
    name = 'migration'
    item = MigrationItem



from dataclasses import dataclass
from logging import getLogger
from typing import List, Optional

import requests
from django.utils.html import escape
from django.utils.safestring import mark_safe

from pytest_logikal.docker import Service

logger = getLogger(__name__)


@dataclass
class ValidationError:
    message: str
    severity: str
    extract: Optional[str]
    first_line: int
    last_line: int


class Validator:
    def __init__(self, service_name: str = 'validator', port: str = '8888/tcp'):
        service = Service(name=service_name, ready_log_text='Checker service started')
        self._validator_url = f'http://127.0.0.1:{service.container_port(port)}'
        logger.debug(
            f'Using HTML/CSS/SVG validator service running at {self._validator_url} '
            f'(container: {service.container.short_id})'
        )

    def errors(self, content: str, content_type: str = 'text/html') -> List[ValidationError]:
        # Checking content
        if not content:
            raise RuntimeError('Empty content')

        response = requests.post(
            self._validator_url, params={'out': 'json'}, headers={'Content-Type': content_type},
            data=content.encode(), timeout=10,
        )
        if response.status_code != 200:
            raise RuntimeError(f'Cannot validate content: {response}')

        # Parsing error messages
        source_lines = content.splitlines()
        errors: List[ValidationError] = []

        for message in response.json()['messages']:
            last_line = message.get('lastLine')
            first_line = message.get('firstLine', last_line)

            if first_line and source_lines[first_line - 1].endswith('<!-- validator: ignore -->'):
                continue

            extract = message.get('extract')
            if extract and (start := message.get('hiliteStart')):
                length = message.get('hiliteLength')
                extract = (
                    escape(extract[:start])
                    + mark_safe('<span class="hll">')  # nosec: fixed literal
                    + escape(extract[start:start + length])
                    + mark_safe('</span>')  # nosec: fixed literal
                    + escape(extract[start + length:])
                )

            errors.append(ValidationError(
                message=message['message'],
                severity=message['type'] if message['type'] != 'info' else 'warning',
                extract=extract,
                first_line=first_line,
                last_line=last_line,
            ))
        return errors



import re
from pathlib import Path

import pytest
from pyorbs.orb import Orb

from pytest_logikal.file_checker import FileCheckItem, FileCheckPlugin
from pytest_logikal.plugin import ItemRunError


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('requirements')
    group.addoption('--requirements', action='store_true', default=False,
                    help='run requirements checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'requirements: checks requirements.')
    if config.option.requirements:
        config.pluginmanager.register(RequirementsPlugin(config=config))


class RequirementsItem(FileCheckItem):
    def runtest(self) -> None:
        if Orb().test(path=self.path, quiet=True):
            relative_path = self.path.relative_to(self.config.invocation_params.dir)
            raise ItemRunError(f'Requirements lockfile "{relative_path}.lock" is outdated')


class RequirementsPlugin(FileCheckPlugin):
    name = 'requirements'
    item = RequirementsItem

    def check_file(self, file_path: Path) -> bool:
        relative_file_path = file_path.relative_to(self.config.invocation_params.dir)
        return (
            bool(re.match(r'reqirements\.txt$|requirements/.+\.txt$', str(relative_file_path)))
            and file_path.with_suffix(file_path.suffix + '.lock').exists()
        )



import subprocess
from pathlib import Path

import pytest
from termcolor import colored

from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError
from pytest_logikal.utils import get_ini_option


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('html')
    group.addoption('--html', action='store_true', default=False, help='run html template checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'html: checks HTML templates.')
    if config.option.html:
        config.pluginmanager.register(HTMLTemplatePlugin(config=config))


# Note: we disabled format checking until some issues are resolved
# (see https://github.com/Riverside-Healthcare/djLint/issues/636)
# (see https://github.com/Riverside-Healthcare/djLint/issues/637)
# Note: the related test is also disabled (see tests/pytest_logikal/test_html.py
class HTMLTemplateItem(CachedFileCheckItem):
    # @staticmethod
    # def _color_diff(line: str) -> str:
    #     if line.startswith('@@'):
    #         return colored(line, 'cyan', force_color=True)
    #     if line.startswith('+'):
    #         return colored(line, 'green', force_color=True)
    #     if line.startswith('-'):
    #         return colored(line, 'red', force_color=True)
    #     return line

    def run(self) -> None:
        messages = []
        max_line_length = str(get_ini_option('max_line_length'))
        common_args = [
            str(self.path),
            '--extension', 'html.j',
            '--indent', '2',
            '--profile', 'jinja',
            '--max-line-length', max_line_length,
            '--max-attribute-length', max_line_length,
            '--linter-output-format', '{line}: error: {message} ({code})',
        ]

        # Check formatting
        # command = ['djlint', '--check', '--preserve-blank-lines', *common_args]
        # process = subprocess.run(command, capture_output=True, text=True, check=False)  # nosec
        # if process.returncode:
        #     errors = process.stdout.strip().replace('@@\n\n', '@@\n')
        #     errors = '\n'.join(self._color_diff(line) for line in errors.splitlines()[2:-2])
        #     messages.append(errors or process.stderr.strip())

        # Lint
        ignore = [
            'H023',  # we allow some entity references (e.g. quotes, special spaces, dashes)
            'H031',  # meta keywords are not that useful anymore
            'J004', 'J018',  # we have our own functions for Jinja environments
            'T002',  # we always use single quotes
            'T003',  # we don't mandate named end blocks
        ]
        command = ['djlint', '--lint', '--ignore', ','.join(ignore), *common_args]
        process = subprocess.run(command, capture_output=True, text=True, check=False)  # nosec
        if process.returncode:
            errors = process.stdout.strip()
            errors = '\n'.join(errors.splitlines()[2:-2])
            messages.append(errors or process.stderr.strip())

        # Report errors
        if messages:
            separator = colored('Errors:', 'red', attrs=['bold'], force_color=True)
            raise ItemRunError(f'\n\n{separator}\n'.join(messages))


class HTMLTemplatePlugin(CachedFileCheckPlugin):
    name = 'html'
    item = HTMLTemplateItem

    def check_file(self, file_path: Path) -> bool:
        return str(file_path).endswith('.html.j')



import logging
import os
import shutil
import subprocess
import sys
from importlib.util import find_spec
from itertools import chain
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union

import pytest
import tomli
from termcolor import colored

sys.path.insert(0, os.getcwd())

PYPROJECT = (
    tomli.loads(Path('pyproject.toml').read_text(encoding='utf-8'))
    if Path('pyproject.toml').exists() else {}
)
PLUGINS = {
    'core': [
        'mypy', 'bandit', 'build', 'docs', 'isort', 'licenses', 'pylint', 'requirements', 'style',
    ],
    'django': ['migration', 'translations', 'html', 'css', 'svg', 'js'],
}
DEFAULT_INI_OPTIONS: Dict[str, Any] = {
    'max_line_length': {'value': 99, 'help': 'the maximum line length to use'},
    'max_complexity': {'value': 10, 'help': 'the maximum complexity to allow'},
    'cov_fail_under': {'value': 100, 'help': 'target coverage percentage'},
}
EXTRAS = {
    'browser': bool(find_spec('selenium')),
    'django': bool(find_spec('pytest_django')),
}

ReportInfoType = Tuple[Union[os.PathLike, str], Optional[int], str]


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('logikal')
    group.addoption('--live', action='store_true', help='run live tests')
    group.addoption('--fast', action='store_true', help='do not run any additional checks')
    group.addoption('--clear', action='store_true', help='clear cache before running tests')
    group.addoption('--no-defaults', action='store_true', help='do not use our own defaults')
    group.addoption('--no-mypy', action='store_true', help='do not use mypy')
    group.addoption('--no-bandit', action='store_true', help='do not use bandit')
    group.addoption('--no-build', action='store_true', help='do not run build checks')
    group.addoption('--no-docs', action='store_true', help='do not run documentation checks')
    group.addoption('--no-isort', action='store_true', help='do not use isort')
    group.addoption('--no-licenses', action='store_true', help='do not check licenses')
    group.addoption('--no-pylint', action='store_true', help='do not use pylint')
    group.addoption('--no-requirements', action='store_true', help='do not check requirements')
    group.addoption('--no-style', action='store_true', help='do not use pycodestyle & pydocstyle')
    if EXTRAS['django']:
        group.addoption('--no-migration', action='store_true', help='do not check migrations')
        group.addoption('--no-translations', action='store_true', help='do not check translations')
        group.addoption('--no-html', action='store_true', help='do not run html template checks')
        group.addoption('--no-css', action='store_true', help='do not run css checks')
        group.addoption('--no-svg', action='store_true', help='do not run svg checks')
        group.addoption('--no-js', action='store_true', help='do not run js checks')
        group.addoption('--no-install', action='store_true', help='do not install packages')

    for option, entry in DEFAULT_INI_OPTIONS.items():
        parser.addini(option, default=str(entry['value']), help=entry['help'])


def pytest_addhooks(pluginmanager: pytest.PytestPluginManager) -> None:
    for extra, installed in EXTRAS.items():
        if not installed:
            pluginmanager.set_blocked(f'logikal_{extra}')
            for plugin in PLUGINS.get(extra, [extra]):
                pluginmanager.set_blocked(f'logikal_{plugin}')


# Untyped decorator (see https://github.com/pytest-dev/pytest/issues/7469)
@pytest.hookimpl(hookwrapper=True)  # type: ignore[misc]
def pytest_load_initial_conftests(early_config: pytest.Config, args: List[str]) -> Iterator[None]:
    if '--no-defaults' in args:
        yield

    # Updating arguments
    args.extend(['--strict-config', '--strict-markers'])
    namespace = early_config.known_args_namespace
    if '--live' in args:
        args.append('--capture=no')
        namespace.capture = 'no'
    if '--clear' in args:
        args.append('--cache-clear')
        namespace.cacheclear = True
    if '-r' not in args:
        args.extend(['-r', 'fExX'])
    if '-n' not in args:
        args.extend(['-n', 'auto' if '--live' not in args else '0'])
        namespace.dist = 'load' if '--live' not in args else 'no'
    for plugin in chain(*(plugins for name, plugins in PLUGINS.items() if EXTRAS.get(name, True))):
        if '--fast' not in args and f'--no-{plugin}' not in args:
            args.append(f'--{plugin}')
            setattr(namespace, plugin, True)
    if all(arg not in args for arg in ['--cov', '--no-cov', '--live', '--fast']):
        args.extend(['--cov', '--no-cov-on-fail'])
        namespace.no_cov_on_fail = True

    # Setting coverage options
    if '--cov' in args:
        fail_under = early_config.getini('cov_fail_under')
        namespace.cov_source = ['.']
        namespace.cov_fail_under = float(fail_under) if '.' in fail_under else int(fail_under)
        namespace.cov_report = {'term-missing': 'skip-covered'}

    # Updating config
    ini_defaults: Dict[str, Union[str, List[str]]] = {
        'console_output_style': 'classic',
        'xfail_strict': 'True',
        'filterwarnings': ['error'],
        'log_cli': 'True' if '--live' in args else 'False',
        'log_level': 'DEBUG',
        'log_format': '%(asctime)s.%(msecs)03d %(levelname)s %(message)s (%(name)s:%(lineno)s)',
        'log_date_format': '%Y-%m-%d %H:%M:%S',
        'log_auto_indent': 'True',
    }
    early_config.inicfg = {**ini_defaults, **early_config.inicfg}
    yield


def install_packages(node_prefix: Optional[Path] = None) -> None:
    node_prefix = node_prefix or Path(__file__).parent
    if not (node_prefix / 'node_modules').exists():
        print(colored('Installing Node.js packages', attrs=['bold']))
        command = ['npm', 'install', '--no-save', '--prefix', str(node_prefix)]
        subprocess.run(command, text=True, check=True)  # nosec: secure, not using untrusted input


def pytest_configure(config: pytest.Config) -> None:
    # Installing Node.js packages
    if EXTRAS['django'] and not config.getoption('no_install'):
        install_packages()

    # Clearing cache
    if config.getoption('clear'):
        print(colored('Clearing cache', 'yellow', attrs=['bold']))
        Path('.coverage').unlink(missing_ok=True)
        shutil.rmtree('.pytest_cache', ignore_errors=True)
        shutil.rmtree('.mypy_cache', ignore_errors=True)

    # Hiding information
    if not config.getoption('verbose'):
        # Hiding overly verbose debug and info log messages
        logging.getLogger('django_migration_linter').setLevel(logging.WARNING)
        logging.getLogger('docker').setLevel(logging.INFO)
        logging.getLogger('faker').setLevel(logging.INFO)
        logging.getLogger('filelock').setLevel(logging.WARNING)
        logging.getLogger('matplotlib').setLevel(logging.INFO)
        logging.getLogger('PIL').setLevel(logging.INFO)
        logging.getLogger('pydocstyle').setLevel(logging.WARNING)
        logging.getLogger('pytest_logikal.validator').setLevel(logging.INFO)
        logging.getLogger('urllib3').setLevel(logging.INFO)
        logging.getLogger('pydocstyle').setLevel(logging.ERROR)

        # Hiding worker information lines
        config.pluginmanager.get_plugin('reports').getworkerinfoline = lambda *_args, **_kwargs: ''

    # Configuring mypy
    mypy = config.pluginmanager.get_plugin('mypy')
    mypy.mypy_argv = [
        '--no-error-summary',
        '--strict',
        '--show-column-numbers',
        '--warn-unreachable',
    ]


def unified_reportinfo(item: pytest.Item, header: str) -> Callable[[], ReportInfoType]:
    def reportinfo() -> ReportInfoType:
        return (item.path, None, header)
    return reportinfo


# Untyped decorator (see https://github.com/pytest-dev/pytest/issues/7469)
@pytest.hookimpl(trylast=True)  # type: ignore[misc]
def pytest_collection_modifyitems(items: List[pytest.Item]) -> None:
    for item in items:
        source = item.nodeid.split('::')[-1].lower()
        if source == 'mypy':
            header = f'[{item.name}] {item.path.relative_to(item.config.invocation_params.dir)}'
            setattr(item, 'reportinfo', unified_reportinfo(item=item, header=header))
        if source == 'mypy-status':
            setattr(item, '_nodeid', '::check::mypy-status')
            setattr(item, 'reportinfo', unified_reportinfo(item=item, header='[mypy-status]'))



import json
import subprocess
from pathlib import Path

import pytest

from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError
from pytest_logikal.utils import get_ini_option, render_template
from pytest_logikal.validator import Validator


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('css')
    group.addoption('--css', action='store_true', default=False, help='run css checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'css: checks CSS files.')
    if config.option.css:
        config.pluginmanager.register(CSSPlugin(config=config))


class CSSItem(CachedFileCheckItem):
    plugin: 'CSSPlugin'

    def run(self) -> None:
        # Validate
        content = self.path.read_text(encoding='utf-8')
        errors = self.plugin.validator.errors(content, content_type='text/css')
        messages = [
            f'{error.first_line}: validation {error.severity}: {error.message}'
            for error in errors
        ]

        # Lint
        # Note: we cannot specify max_line_length via CLI arguments currently
        # (see https://github.com/stylelint/stylelint/issues/6805)
        context = {'max_line_length': get_ini_option('max_line_length')}
        with render_template(Path(__file__).parent / 'css_config.yml', context) as config_path:
            command = [
                'npx', '--no',
                'stylelint', str(self.path), '--formatter', 'json',
                '--config', str(config_path),
            ]
            process = subprocess.run(  # nosec
                command, capture_output=True, text=True, check=False, cwd=Path(__file__).parent,
            )
        try:
            report = json.loads(process.stdout)[0]
            if report['errored']:
                messages.extend(
                    f'{error["line"]}:{error["column"]}: {error["severity"]}: {error["text"]}'
                    for error in report['warnings']
                )
        except json.decoder.JSONDecodeError:
            messages.append(process.stdout or process.stderr)

        # Report errors
        if messages:
            raise ItemRunError('\n'.join(messages))


class CSSPlugin(CachedFileCheckPlugin):
    name = 'css'
    item = CSSItem

    def __init__(self, config: pytest.Config):
        super().__init__(config=config)
        self.validator = Validator()

    def check_file(self, file_path: Path) -> bool:
        return file_path.suffix == '.css'



import json
import subprocess
from typing import List

import pytest

from pytest_logikal.core import PYPROJECT, ReportInfoType
from pytest_logikal.plugin import Item, ItemRunError, Plugin

ALLOWED_LICENSES = [
    '3-Clause BSD License',
    'Apache 2.0',
    'Apache License 2.0',
    'Apache Software License',
    'BSD 3-Clause',
    'BSD License',
    'BSD',
    'BSD-3-Clause',
    'GNU Lesser General Public License v2 (LGPLv2)',
    'GNU Lesser General Public License v2 or later (LGPLv2+)',
    'GNU Lesser General Public License v3 (LGPLv3)',
    'GNU Library or Lesser General Public License (LGPL)',
    'ISC License (ISCL)',
    'ISC',
    'MIT License',
    'MIT No Attribution License (MIT-0)',
    'MIT',
    'Mozilla Public License 2.0 (MPL 2.0)',
    'Public Domain',
    'Python Software Foundation License',
    'The MIT License (MIT)',
    'The Unlicense (Unlicense)',
]

ALLOWED_PACKAGES = {
    'facebook-business': 'LICENSE.txt',  # only used as a connector
    'djlint': 'GNU General Public License v3 or later (GPLv3+)',  # only used as a local tool
    'html-tag-names': 'GNU General Public License v3 or later (GPLv3+)',  # djlint dependency
    'html-void-elements': 'GNU General Public License v3 or later (GPLv3+)',  # djlint dependency
    'pkg_resources': 'UNKNOWN',  # caused by an Ubuntu bug, see [1]
    'pkg-resources': 'UNKNOWN',  # caused by an Ubuntu bug, see [1]
    'Pillow': 'Historical Permission Notice and Disclaimer (HPND)',  # license is BSD-like
    'pylint': 'GNU General Public License v2 (GPLv2)',  # only used as a local tool
    'pylint-django': 'GPLv2',  # only used as a local tool plugin
    'pylint-plugin-utils': 'GNU General Public License v2 or later (GPLv2+)',  # local tool plugin
    'python-lsp-jsonrpc': 'UNKNOWN',  # license is MIT
}
# [1] https://bugs.launchpad.net/ubuntu/+source/python-pip/+bug/1635463


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('licenses')
    group.addoption('--licenses', action='store_true', default=False, help='run license checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'licenses: checks licenses.')
    if config.option.licenses:
        config.pluginmanager.register(LicensePlugin(config=config))


class LicenseItem(Item):
    def runtest(self) -> None:
        command = ['pip-licenses', '--format=json', '--with-system', '--with-urls']

        # This subprocess call is secure as it is not using untrusted input
        process = subprocess.run(command, capture_output=True, text=True, check=True)  # nosec
        packages = json.loads(process.stdout)

        pyproject_licenses = PYPROJECT.get('tool', {}).get('licenses', {})
        allowed_licenses = pyproject_licenses.get('allowed_licenses', ALLOWED_LICENSES)
        allowed_packages = pyproject_licenses.get('allowed_packages', ALLOWED_PACKAGES)
        allowed_licenses.extend(pyproject_licenses.get('extend_allowed_licenses', []))
        allowed_packages.update(pyproject_licenses.get('extend_allowed_packages', {}))

        warnings: List[str] = []
        for package in sorted(packages, key=lambda item: (item['License'], item['Name'].lower())):
            name, version, url = package['Name'], package.get('Version'), package.get('URL')
            licenses = [license.strip() for license in package['License'].split(';')]
            allowed_package_license = allowed_packages.get(name) in licenses
            allowed_license = any(license in allowed_licenses for license in licenses)
            if (not allowed_package_license and not allowed_license):
                url = f' ({url})' if url not in ('UNKNOWN', None) else ''
                warnings.append(f'Warning: {package["License"]}: {name}=={version}{url}')
        if warnings:
            raise ItemRunError('\n'.join(warnings))

    def reportinfo(self) -> ReportInfoType:
        return (self.path, None, f'[{LicensePlugin.name}]')


class LicensePlugin(Plugin):
    name = 'licenses'
    item = LicenseItem



import json
import subprocess
from pathlib import Path

import pytest

from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError
from pytest_logikal.utils import get_ini_option


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('js')
    group.addoption('--js', action='store_true', default=False, help='run js checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'js: checks JS files.')
    if config.option.js:
        config.pluginmanager.register(JSPlugin(config=config))


class JSItem(CachedFileCheckItem):
    def run(self) -> None:
        command = [
            'npx', '--no',
            'eslint', str(self.path), '--format', 'json', '--max-warnings', '0',
            '--config', str(Path(__file__).parent / 'js_config.yml'),
            '--rule', f'max-len: [error, {get_ini_option("max_line_length")}]',
            '--rule', f'complexity: [error, {get_ini_option("max_complexity")}]',
        ]
        process = subprocess.run(  # nosec
            command, capture_output=True, text=True, check=False, cwd=Path(__file__).parent,
        )
        try:
            severity = {1: 'warning', 2: 'error'}
            report = json.loads(process.stdout)[0]
            if report['messages']:
                raise ItemRunError('\n'.join(
                    f'{error["line"]}:{error["column"]}: {severity[error["severity"]]}: '
                    + f'{error["message"]}'
                    + (f' ({error["ruleId"]})' if error['ruleId'] else '')
                    for error in report['messages']
                ))
        except json.decoder.JSONDecodeError as error:
            raise ItemRunError((process.stdout or process.stderr).strip()) from error


class JSPlugin(CachedFileCheckPlugin):
    name = 'js'
    item = JSItem

    def check_file(self, file_path: Path) -> bool:
        return file_path.suffix == '.js'



import os
import subprocess
from time import sleep, time
from typing import Optional

from docker import DockerClient, from_env as client_from_env
from docker.models.containers import Container
from termcolor import colored

from pytest_logikal.core import PYPROJECT


class Service:
    def __init__(  # pylint: disable=too-many-arguments
        self,
        name: str,
        project: Optional[str] = None,
        start_timeout_seconds: float = 30,
        ready_log_text: Optional[str] = None,
        log_poll_seconds: float = 3,
    ):
        self.name = name
        self.project = project or PYPROJECT['project']['name']
        self.start_timeout_seconds = start_timeout_seconds
        self.ready_log_text = ready_log_text
        self.log_poll_seconds = log_poll_seconds
        try:
            client = client_from_env()
            self.container = self._running_container(client=client, start_services=True)
        finally:
            client.close()

    def __str__(self) -> str:
        return f'Docker Compose project "{self.project}" service "{self.name}"'

    def start_services(self) -> None:
        print(colored('Starting Docker Compose services', 'yellow', attrs=['bold']))
        command = [
            'docker', 'compose', 'up', '--detach',
            *(['--quiet-pull'] if 'GITHUB_ACTIONS' in os.environ else []),
            '--wait', '--wait-timeout', str(self.start_timeout_seconds),
        ]
        subprocess.run(command, text=True, check=True)  # nosec: secure, not using untrusted input
        print()

    def _running_container(self, client: DockerClient, start_services: bool) -> Container:
        filters = {
            'status': 'running',
            'label': [
                f'com.docker.compose.project={self.project}',
                f'com.docker.compose.service={self.name}',
            ],
        }
        if containers := client.containers.list(limit=1, filters=filters):
            container = containers[0]
            container.reload()  # refresh container attributes

            health = container.attrs['State'].get('Health', {})
            health_status = health.get('Status')
            if health_status and health_status != 'healthy':
                logs = [log['Output'].strip() for log in health.get('Log', {}) if 'Output' in log]
                logs_str = ('\n\nLogs:\n' + '\n'.join(logs)) if logs else ''
                raise RuntimeError(f'{self} is not healthy{logs_str}')

            if self.ready_log_text and not start_services:
                print(f'Waiting for service "{self.name}" to be ready...')
                wait_start_time = time()
                while (  # pylint: disable=while-used
                    self.ready_log_text not in container.logs().decode()
                ):
                    if time() - wait_start_time >= self.start_timeout_seconds:
                        raise RuntimeError(f'{self} is not ready')
                    sleep(self.log_poll_seconds)
                print()

            return container

        if start_services:
            self.start_services()
            return self._running_container(client=client, start_services=False)

        raise RuntimeError(f'{self} not found')

    def container_port(self, service_port: str) -> str:
        return str(self.container.ports[service_port][0]['HostPort'])



from pathlib import Path

import pytest

from pytest_logikal.file_checker import CachedFileCheckItem, CachedFileCheckPlugin
from pytest_logikal.plugin import ItemRunError
from pytest_logikal.validator import Validator


def pytest_addoption(parser: pytest.Parser) -> None:
    group = parser.getgroup('svg')
    group.addoption('--svg', action='store_true', default=False, help='run svg checks')


def pytest_configure(config: pytest.Config) -> None:
    config.addinivalue_line('markers', 'svg: checks SVG files.')
    if config.option.svg:
        config.pluginmanager.register(SVGPlugin(config=config))


class SVGItem(CachedFileCheckItem):
    plugin: 'SVGPlugin'

    def run(self) -> None:
        content = self.path.read_text(encoding='utf-8')
        errors = self.plugin.validator.errors(content, content_type='image/svg+xml')
        messages = [
            f'{error.first_line}: {error.severity}: {error.message}'
            for error in errors if 'Using the preset for SVG' not in error.message
        ]
        if messages:
            raise ItemRunError('\n'.join(messages))


class SVGPlugin(CachedFileCheckPlugin):
    name = 'svg'
    item = SVGItem

    def __init__(self, config: pytest.Config):
        super().__init__(config=config)
        self.validator = Validator()

    def check_file(self, file_path: Path) -> bool:
        return file_path.suffix == '.svg'



try:
    import zoneinfo  # type: ignore[import]
except ImportError:
    from backports import zoneinfo  # note: this is built-in after Python 3.9+
from typing import Any, Callable, Iterable

import pytest
from django.conf import settings
from django.urls import reverse
from django.utils import timezone as django_timezone, translation
from factory import random as factory_random
from mypy_django_plugin import config as mypy_django_plugin_config
from pytest_django.live_server_helper import LiveServer

from pytest_logikal.utils import Fixture, Function

DEFAULT_RANDOM_SEED = 42
LiveURL = Callable[[str], str]


def pytest_configure(config: pytest.Config) -> None:
    # Patching django-stubs
    def parse_toml_file(self: Any, *_args: Any, **_kwargs: Any) -> None:
        self.django_settings_module = str(config.inicfg['DJANGO_SETTINGS_MODULE'])

    mypy_django_plugin_config.DjangoPluginConfig.parse_toml_file = parse_toml_file  # type: ignore


@pytest.fixture(scope='session', autouse=True)
def faker_seed() -> int:
    """
    Set a default seed for Faker for deterministic testing. Automatically applied.
    """
    return DEFAULT_RANDOM_SEED


@pytest.fixture(scope='function', autouse=True)
def factory_seed() -> None:
    """
    Set a default seed for pytest-factoryboy for deterministic testing. Automatically applied.
    """
    factory_random.reseed_random(DEFAULT_RANDOM_SEED)


@pytest.fixture
def live_url(live_server: LiveServer) -> LiveURL:  # noqa: D400,D402,D415,D417
    """
    live_url(name: str) -> str

    Return the path to a URL.

    Args:
        name: The URL pattern name.

    """
    def live_url_path(name: str) -> str:
        return live_server.url + reverse(name)
    return live_url_path


@pytest.fixture
def language(request: Any) -> Iterable[str]:
    """
    Return the currently activated language.
    """
    if language_code := getattr(request, 'param', None):
        with translation.override(language_code):
            yield language_code
    else:
        yield translation.get_language()


def set_language(*language_codes: str) -> Fixture[Any]:
    """
    Mark a test to run with each of the specified languages.

    Args:
        *language_codes: The language codes to use.

    .. note:: You must also use the :func:`language <pytest_logikal.django.language>` fixture in
        your test when applying this decorator.

    """
    def parametrized_test_function(function: Function) -> Any:
        return pytest.mark.parametrize(
            argnames='language', argvalues=language_codes, indirect=True,
            ids=lambda value: f'language={value}'
        )(function)
    return parametrized_test_function


def all_languages() -> Fixture[Any]:
    """
    Mark a test to run with every available language.

    .. note:: You must also use the :func:`language <pytest_logikal.django.language>` fixture in
        your test when applying this decorator.

    """
    return set_language(*(language_code for language_code, _ in settings.LANGUAGES))


@pytest.fixture
def timezone(request: Any) -> Iterable[str]:
    """
    Return the current time zone ID.
    """
    if zone_id := getattr(request, 'param', None):
        current_timezone = django_timezone.get_current_timezone()
        django_timezone.activate(zoneinfo.ZoneInfo(zone_id))
        yield zone_id
        django_timezone.activate(current_timezone)
    else:
        yield django_timezone.get_current_timezone_name()


def set_timezone(*zone_ids: str) -> Fixture[Any]:
    """
    Mark a test to run with each of the specified time zones.

    Args:
        *zone_ids: The time zone IDs to use.

    .. note:: You must also use the :func:`timezone <pytest_logikal.django.timezone>` fixture in
        your test when applying this decorator.

    """
    def parametrized_test_function(function: Function) -> Any:
        return pytest.mark.parametrize(
            argnames='timezone', argvalues=zone_ids, indirect=True,
            ids=lambda value: f'timezone={value}'
        )(function)
    return parametrized_test_function
